/*
 * Copyright (c) 2004 DotNetGuru and the individuals listed
 * on the ChangeLog entries.
 *
 * Authors :
 *   Thomas Gil (thomas.gil@dotnetguru.org)
 *   Evain Jb   (jb@evain.net)
 *
 * This is a free software distributed under a BSD license
 * See LICENSE.BSD file for more details
 *
 *****************************************************************************/

namespace Ilml.Adapters {

    using System;
    using System.IO;
    using System.Reflection;
    using System.Xml;
    using System.Xml.XPath;

    using Rail.Reflect;

	internal sealed class AssemblyAdapter : XmlAdapter {

        #region singleton implementation

		private AssemblyAdapter() : base("Assembly") {}

		public static readonly AssemblyAdapter Instance = 
			new AssemblyAdapter();

        #endregion

		private RAssemblyDef m_currentAssembly;
        private XPathNavigator m_navigator;

		public RAssemblyDef CurrentAssembly {
			get{ return m_currentAssembly; }
		}

        public XPathNavigator Navigator {
            get { return m_navigator; }
        }

		public void ToXml(RAssemblyDef a, XmlWriter writer) {
			writer.WriteStartDocument();
			writer.WriteStartElement(TagName, Utils.ilNS);

			writer.WriteAttributeString("fullName", Utils.ilNS,
                a.Name.FullName);

			if (a.EntryPoint != null) {
				writer.WriteAttributeString("entryPointSignature",
                    Utils.ilNS, Utils.GetSignature(a.EntryPoint));
			}

            if (a.PublicKey != null) {
                writer.WriteAttributeString("publicKey",
                    Utils.ilNS, Convert.ToBase64String(a.PublicKey));
            }

            AssemblyNameAdapter.Instance.ToXml(a.Name, writer);

			foreach (RAssemblyName asm in a.GetReferencedAssemblies()) {
                writer.WriteStartElement("ReferencedAssembly");
                writer.WriteAttributeString("fullName", Utils.ilNS,
                    asm.FullName);

				AssemblyNameAdapter.Instance.ToXml(asm, writer);
                
                writer.WriteEndElement();
			}

            foreach (RModule m in a.GetModules()) {
                if (m is RModuleRef) {
                    writer.WriteStartElement("ReferencedModule",
                        Utils.ilNS);
                    writer.WriteAttributeString("name", Utils.ilNS, m.Name);
                    writer.WriteAttributeString("fileName",
                        Utils.ilNS, m.FileName);
                    writer.WriteEndElement();
                }
            }

            foreach (RCustomAttrib attr in a.GetCustomAttributes(false)) {
                AttributeAdapter.Instance.ToXml(attr, writer);
            }

            foreach (RModule m in a .GetModules()) {
                if (m is RModuleDef) {
                    ModuleAdapter.Instance.ToXml(m, writer);
                }
            }

            foreach (RResource r in a.GetResources()) {
                ResourceAdapter.Instance.ToXml(r, writer);
            }

			writer.WriteEndElement();
			writer.WriteEndDocument();
		}

        public RAssemblyDef FromXml(XPathNavigator nav) {
            m_navigator = nav;

            XPathNodeIterator it = nav.Select(
                AssemblyNameAdapter.Instance.TagName);
            it.MoveNext();
            
            RAssemblyDef asm = new RAssemblyDef(
                AssemblyNameAdapter.Instance.FromXml(it.Current));
            m_currentAssembly = asm;

            if (Utils.IsAttributePresent(nav, "publicKey")) {
                asm.PublicKey = Convert.FromBase64String(
                    nav.GetAttribute("publicKey", Utils.ilNS));
            }

            it = nav.Select("ReferencedAssembly");
            while (it.MoveNext()) {
                XPathNodeIterator refasm = it.Current.Select(
                    AssemblyNameAdapter.Instance.TagName);
                refasm.MoveNext();
                asm.AddReferencedAssembly(AssemblyNameAdapter.Instance.FromXml(
                    refasm.Current));
            }

            it = nav.Select("ReferencedModule");
            while (it.MoveNext()) {
                asm.DefineModuleRef(
                    it.Current.GetAttribute("name", Utils.ilNS),
                    it.Current.GetAttribute("fileName", Utils.ilNS));
            }

            it = nav.Select(AttributeAdapter.Instance.TagName);
            while (it.MoveNext()) {
                AttributeAdapter.Instance.FromXml(asm, it.Current);
            }

            it = nav.Select(ModuleAdapter.Instance.TagName);
            while (it.MoveNext()) {
                ModuleAdapter.Instance.FromXml(asm, it.Current);
            }

            it = nav.Select(ResourceAdapter.Instance.TagName);
            while (it.MoveNext()) {
                ResourceAdapter.Instance.FromXml(asm, it.Current);
            }

            if (Utils.IsAttributePresent(nav, "entryPointSignature")) {
                asm.EntryPoint = MethodAdapter.Instance.GetMethod(
                    nav.GetAttribute("entryPointSignature", Utils.ilNS));
            }
            return asm;
        }
	}
}
