/*
 * Copyright (c) 2004 DotNetGuru and the individuals listed
 * on the ChangeLog entries.
 *
 * Authors :
 *   Evain Jb   (jb@evain.net)
 *
 * This is a free software distributed under a BSD license
 * See LICENSE.BSD file for more details
 *
 *****************************************************************************/

namespace Ilml.Adapters {

    using System;
    using System.Collections;
    using System.Reflection.Emit;
    using System.Xml.XPath;

    using Rail.MSIL;
    using Rail.Reflect;

    #region CodeTrees

    internal abstract class CodeTree {
        protected CodeNode m_root;

        public CodeNode Root {
            get { return m_root; }
        }

        protected abstract void Parse();
    }

    #region CodeTreeReflect

    internal sealed class CodeTreeReflect : CodeTree {
        private Code m_code;
        private ExceptionTable m_table;

        private IList m_exceptions;

        private Stack m_exBlocks;
        private CodeNode m_last;

        public CodeTreeReflect(MethodBody mb) {
            mb.Code.updateInstructionIndex();
            this.m_code = mb.Code;
            this.m_table = mb.ExceptionTable;
            this.m_exceptions = new ArrayList();
            this.m_exBlocks = new Stack();
            this.Parse();
        }

        protected override void Parse() {
            if (m_table != null) {
                BuildExceptions();
            }
            BuildTree();
        }

        private bool IsBlockBuilded(int start, int end) {
            foreach (CodeNodeException cne in m_exceptions) {
                if (cne.StartOffset == start && cne.EndOffset == end) {
                    return true;
                }
            }
            return false;
        }

        private CodeNodeException GetNextExceptionBlock() {
            CodeNodeException ret = null;
            if (m_exceptions.Count > 0) {
                ret = (CodeNodeException)m_exceptions[0];
                for (int i = 1 ; i < m_exceptions.Count ; i++) {
                    CodeNodeException cne = (CodeNodeException)m_exceptions[i];
                    if (cne.StartOffset <= ret.StartOffset &&
                        cne.EndOffset > ret.EndOffset) {

                        ret = cne;
                    }
                }
                m_exceptions.Remove(ret);
            }
            return ret;
        }

        private void BuildExceptions() {
            for (int i = 0 ; i < m_table.Count() ; i++) {
                ExceptionBlock eb = m_table[i];
                if (!IsBlockBuilded(eb.StartOfProtectedScope.Index,
                    eb.EndOfProtectedScope.Index)) {

                    m_exceptions.Add(new CodeNodeException(
                        this,
                        CodeNodeExceptionType.Try,
                        eb.StartOfProtectedScope.Index,
                        eb.EndOfProtectedScope.Index));
                }
                if (eb.GetCatchHandlerCount() > 0) {
                    for (int j = 0 ; j < eb.GetCatchHandlerCount() ; j++) {
                        ExceptionHandler eh = eb.GetCatchHandler(j);
                        if (!IsBlockBuilded(eh.StartInstruction.Index,
                            eh.EndInstruction.Index)) {

                            CodeNodeException cne = new CodeNodeException(
                                this,
                                CodeNodeExceptionType.Catch,
                                eh.StartInstruction.Index,
                                eh.EndInstruction.Index);
                            cne.CatchType = eh.ExceptionType;
                            m_exceptions.Add(cne);
                        }
                    }
                }
                if (eb.FaultHandler != null) {
                    ExceptionHandler eh = eb.FaultHandler;
                    if (!IsBlockBuilded(eh.StartInstruction.Index,
                        eh.EndInstruction.Index)) {

                        m_exceptions.Add(new CodeNodeException(
                            this,
                            CodeNodeExceptionType.Fault,
                            eh.StartInstruction.Index,
                            eh.EndInstruction.Index));
                    }
                }
                if (eb.FinallyHandler != null) {
                    ExceptionHandler eh = eb.FinallyHandler;
                    if (!IsBlockBuilded(eh.StartInstruction.Index,
                        eh.EndInstruction.Index)) {

                        m_exceptions.Add(new CodeNodeException(
                            this,
                            CodeNodeExceptionType.Finally,
                            eh.StartInstruction.Index,
                            eh.EndInstruction.Index));
                    }
                }
            }
        }

        private void BuildTree() {
            for (int i = 0 ; i < m_code.InstructionCount ; i++) {
                BuildSequence(i);
            }
        }

        private bool ExceptionStartsAt(int pos) {
            foreach (CodeNodeException cne in m_exceptions) {
                if (cne.StartOffset == pos) {
                    return true;
                }
            }
            return false;
        }

        private bool ExceptionEndsAt(int pos) {
            foreach (CodeNodeException cne in m_exBlocks) {
                if (cne.EndOffset == pos + 1) {
                    return true;
                }
            }
            return false;
        }

        private void BuildSequence(int seq) {
            while (ExceptionStartsAt(seq)) {
                PushException(GetNextExceptionBlock());
            }
            AddInstruction(m_code[seq]);
            while (ExceptionEndsAt(seq)) {
                PopException();
            }
        }

        private void AddNode(CodeNode node) {
            if (m_last is CodeNodeInstruction) {
                node.Previous = m_last;
                m_last.Next = node;
                m_last = node;
            } else if (m_last is CodeNodeException) {
                CodeNodeException cne = m_last as CodeNodeException;
                if (cne.First == null) {
                    cne.First = node;
                } else {
                    m_last.Next = node;
                    node.Previous = m_last;
                }
                m_last = node;
            } else {
                m_root = m_last = node;
            }
        }

        private void PushException(CodeNodeException cne) {
            m_exBlocks.Push(cne);
            AddNode(cne);
        }

        private void AddInstruction(Instruction i) {
            CodeNodeInstruction cni = new CodeNodeInstruction(this, i);
            AddNode(cni);
        }

        private void PopException() {
            CodeNodeException cne = (CodeNodeException)m_exBlocks.Pop();
            m_last = cne;
        }
    }

    #endregion

    #region CodeTreeXml

    internal sealed class CodeTreeXml : CodeTree {
        private XPathNavigator m_nav;
        private IList m_instructions;
        private int m_index;
        private string level = string.Format(
            "node()[local-name()='{0}' or local-name()='{1}']",
            InstructionAdapter.Instance.TagName,
            ExceptionAdapter.Instance.TagName);

        public int IndexedInstruction {
            get { return m_index; }
        }
        
        public CodeTreeXml(XPathNavigator method) {
            this.m_nav = method;
            this.m_instructions = new ArrayList();
            this.Parse();
        }

        protected override void Parse() {
            m_root = Parse(m_nav.Select(level));
            IndexInstructions(m_root);
            IndexExceptions(m_root);
        }

        private CodeNode Parse(XPathNodeIterator it) {
            CodeNode node = null, curs = null;
            while (it.MoveNext()) {
                if (it.Current.Name == InstructionAdapter.Instance.TagName) {
                    CodeNodeInstruction ins = GetInstruction(it.Current);
                    m_instructions.Add(ins);
                    curs = ins;
                } else if (it.Current.Name ==
                    ExceptionAdapter.Instance.TagName) {

                    CodeNodeException ex = GetException(it.Current);
                    ex.First = Parse(it.Current.Select(level));
                    curs = ex;
                }
                if (node == null) {
                    node = curs;
                } else {
                    node.Next = curs;
                    curs.Previous = node;
                    node = node.Next;
                }
            }
            return node.GetHead();
        }

        private string GetStartIdentifier(XPathNavigator nav) {
            XPathNodeIterator it = nav.Select(string.Format(
                "descendant::{0}[position()=1]",
                InstructionAdapter.Instance.TagName));
            it.MoveNext();
            return it.Current.GetAttribute("offset", Utils.ilNS);
        }

        private string GetEndIdentifier(XPathNavigator nav) {
            XPathNodeIterator it = nav.Select(string.Format(
                "following::{0}[position()=1]",
                InstructionAdapter.Instance.TagName));
            it.MoveNext();
            return it.Current.GetAttribute("offset", Utils.ilNS);
        }

        private CodeNodeException GetException(XPathNavigator nav) {
            CodeNodeExceptionType t = (CodeNodeExceptionType)Enum.Parse(
                typeof(CodeNodeExceptionType),
                nav.GetAttribute("type", Utils.ilNS), true);
            CodeNodeException ex = new CodeNodeExceptionXml(this, t,
                GetStartIdentifier(nav), GetEndIdentifier(nav));
            if (t == CodeNodeExceptionType.Catch) {
                ex.CatchType = TypeAdapter.Instance.GetType(
                    nav.GetAttribute("catchType", Utils.ilNS));
            }
            return ex;
        }

        public Instruction GetInstruction(int offset) {
            Instruction ret = null;
            if (offset < m_instructions.Count) {
                CodeNodeInstruction ins = 
                    (CodeNodeInstruction)m_instructions[offset];
                if (ins != null) {
                    ret = ins.Instruction;
                }
            }
            return ret;
        }

        public int GetIndexForIdentifier(string id) {
            for (int i = 0 ; i < m_instructions.Count ; i++) {
                CodeNodeInstructionXml ins = 
                    (CodeNodeInstructionXml)m_instructions[i];
                if (ins.Identifier == id) {
                    return ins.Offset;
                }
            }
            return -1;
        }

        private CodeNodeInstruction GetInstruction(XPathNavigator nav) {
            Instruction i = CILFactory.Instance.GetInstruction(
                nav.GetAttribute("code", Utils.ilNS));

            CodeNodeInstruction ins = new CodeNodeInstructionXml(this, i,
                nav.GetAttribute("offset", Utils.ilNS));
            if (i.OpCode.OperandType == OperandType.InlineString) {
                ins.Operand = nav.GetAttribute("operand", Utils.ilNS);
            } else {
                if (Utils.IsAttributePresent(nav, "operand")) {
                    ins.Operand = nav.GetAttribute("operand", Utils.ilNS);
                }
            }
            return ins;
        }

        private void IndexInstructions(CodeNode node) {
            if (node != null) {
                if (node is CodeNodeInstruction) {
                    CodeNodeInstruction ins = (CodeNodeInstruction)node;
                    ins.Offset = m_index++;
                    IndexInstructions(ins.Next);
                } else if (node is CodeNodeException) {
                    IndexInstructions(((CodeNodeException)node).First);
                    IndexInstructions(node.Next);
                }
            }
        }

        private void IndexExceptions(CodeNode node) {
            if (node != null) {
                if (node is CodeNodeInstruction) {
                    IndexExceptions(node.Next);
                } else if (node is CodeNodeExceptionXml) {
                    CodeNodeExceptionXml ex = (CodeNodeExceptionXml)node;
                    ex.StartOffset = GetIndexForIdentifier(ex.StartIdentifier);
                    ex.EndOffset = GetIndexForIdentifier(ex.EndIdentifier);
                    IndexExceptions(ex.First);
                    IndexExceptions(ex.Next);
                }
            }
        }
    }

    #endregion

    #endregion

    #region CodeNodes

    internal abstract class CodeNode {
        private CodeTree m_tree;
        private CodeNode m_previous;
        private CodeNode m_next;

        protected CodeNode(CodeTree tree) {
            m_tree = tree;
        }

        public CodeTree Tree {
            get { return m_tree; }
        }

        public CodeNode Previous {
            get { return m_previous; }
            set { m_previous = value; }
        }

        public CodeNode Next {
            get { return m_next; }
            set { m_next = value; }
        }

        public CodeNode GetTail() {
            if (m_next != null) {
                return m_next.GetTail();
            } else {
                return this;
            }
        }

        public CodeNode GetHead() {
            if (m_previous != null) {
                return m_previous.GetHead();
            } else {
                return this;
            }
        }
    }

    internal class CodeNodeInstruction : CodeNode {
        private int m_pos;
        private Instruction m_ins;
        private string m_operand;

        public int Offset {
            get { return m_pos; }
            set { m_pos = value; }
        }

        public Instruction Instruction {
            get { return m_ins; }
        }

        public string Operand {
            get { return m_operand; }
            set { m_operand = value; }
        }

        public CodeNodeInstruction(CodeTree tree, Instruction ins)
            : base(tree) {

            this.m_ins = ins;
            this.m_pos = m_ins.Index;
        }
    }

    internal sealed class CodeNodeInstructionXml : CodeNodeInstruction {
        private string m_identifier;

        public string Identifier {
            get { return m_identifier; }
        }

        public CodeNodeInstructionXml(CodeTree tree, Instruction ins,
            string identifier) : base(tree, ins) {

            m_identifier = identifier;
        }
    }

    internal enum CodeNodeExceptionType : byte {
        Try,
        Catch,
        Finally,
        Fault
    }

    internal class CodeNodeException : CodeNode {
        private int m_start, m_end;
        private CodeNodeException m_parent;
        private CodeNode m_first;
        private CodeNodeExceptionType m_nodeType;
        private RType m_type;

        public int StartOffset {
            get { return m_start; }
            set { m_start = value; }
        }

        public int EndOffset {
            get { return m_end; }
            set { m_end = value; }
        }

        public CodeNodeException Parent {
            get { return m_parent; }
            set { m_parent = value; }
        }

        public CodeNode First {
            get { return m_first; }
            set { m_first = value; }
        }

        public CodeNodeExceptionType Type {
            get { return m_nodeType; }
        }

        public RType CatchType {
            get { return m_type; }
            set { m_type = value; }
        }

        public CodeNodeException(CodeTree tree, CodeNodeExceptionType type,
            int start, int end) : base(tree) {

            this.m_nodeType = type;
            this.m_start = start;
            this.m_end = end;
        }

        public CodeNodeException GetTryNode() {
            if (m_nodeType == CodeNodeExceptionType.Try) {
                return this;
            } else {
                return ((CodeNodeException)this.Previous).GetTryNode();
            }
        }
    }

    internal class CodeNodeExceptionXml : CodeNodeException {
        private string m_startId;
        private string m_enId;

        public string StartIdentifier {
            get { return m_startId; }
        }

        public string EndIdentifier {
            get { return m_enId; }
        }

        public CodeNodeExceptionXml(CodeTree tree, CodeNodeExceptionType type,
            string startId, string endId) : base(tree, type, 0, 0) {

            m_startId = startId;
            m_enId = endId;
        }
    }

    #endregion
}

/*
 * Not proud of this
 * But sometimes it works...
 */