/*
 * Copyright (c) 2004 DotNetGuru and the individuals listed
 * on the ChangeLog entries.
 *
 * Authors :
 *   Thomas Gil (thomas.gil@dotnetguru.org)
 *   Evain Jb   (jb@evain.net)
 *
 * This is a free software distributed under a BSD license
 * See LICENSE.BSD file for more details
 *
 *****************************************************************************/

namespace Ilml.Adapters {

    using System;
    using System.Collections;
    using System.Reflection;
    using System.Xml;
    using System.Xml.XPath;

    using Rail.Reflect;

    internal sealed class FieldAdapter : XmlAdapter {

        #region singleton implementation

        private FieldAdapter() : base("Field") {
            m_fields = new Hashtable();
        }
        public static readonly FieldAdapter Instance = 
            new FieldAdapter();

        #endregion

        private IDictionary m_fields;

        public void ToXml(RFieldDef f, XmlWriter writer) {
            writer.WriteStartElement(TagName, Utils.ilNS);
            writer.WriteAttributeString("name", Utils.ilNS, f.Name);
            writer.WriteAttributeString("type", Utils.ilNS,
                Utils.GetFullyQualifiedName(f.FieldType));
            writer.WriteAttributeString("definition", Utils.ilNS,
                Utils.GetDefinition(f));

            if (f.Value != null) {
                writer.WriteAttributeString("value",
                    Utils.ilNS, f.Value.ToString());
            }

            #region FieldAttributes writing

            bool b;

            // Field access attributes
            int mask = (int)(f.Attributes & FieldAttributes.FieldAccessMask);
            b = mask == (int)FieldAttributes.PrivateScope;
            writer.WriteAttributeString("isPrivateScope", Utils.ilNS,
                b.ToString());
            b = mask == (int)FieldAttributes.Private;
            writer.WriteAttributeString("isPrivate", Utils.ilNS,
                b.ToString());
            b = mask == (int)FieldAttributes.FamANDAssem;
            writer.WriteAttributeString("isFamANDAssem", Utils.ilNS,
                b.ToString());
            b = mask == (int)FieldAttributes.Assembly;
            writer.WriteAttributeString("isAssembly", Utils.ilNS,
                b.ToString());
            b = mask == (int)FieldAttributes.FamORAssem;
            writer.WriteAttributeString("isFamORAssem", Utils.ilNS,
                b.ToString());
            b = mask == (int)FieldAttributes.Public;
            writer.WriteAttributeString("isPublic", Utils.ilNS,
                b.ToString());

            // Field contract attributes
            b = (f.Attributes & FieldAttributes.Static) ==
                FieldAttributes.Static;
            writer.WriteAttributeString("isStatic", Utils.ilNS,
                b.ToString());
            b = (f.Attributes & FieldAttributes.InitOnly) ==
                FieldAttributes.InitOnly;
            writer.WriteAttributeString("isInitOnly", Utils.ilNS,
                b.ToString());
            b = (f.Attributes & FieldAttributes.Literal) ==
                FieldAttributes.Literal;
            writer.WriteAttributeString("isLiteral", Utils.ilNS,
                b.ToString());
            b = (f.Attributes & FieldAttributes.NotSerialized) ==
                FieldAttributes.NotSerialized;
            writer.WriteAttributeString("isNotSerialized", Utils.ilNS,
                b.ToString());

            // Special name attribute
            b = (f.Attributes & FieldAttributes.SpecialName) ==
                FieldAttributes.SpecialName;
            writer.WriteAttributeString("isSpecialName", Utils.ilNS,
                b.ToString());

            // Interop attributes
            b = (f.Attributes & FieldAttributes.PinvokeImpl) ==
                FieldAttributes.PinvokeImpl;
            writer.WriteAttributeString("isPinvokeImpl", Utils.ilNS,
                b.ToString());

            // Runtime reserved attributes
            mask = (int)(f.Attributes & FieldAttributes.ReservedMask);
            b = mask == (int)FieldAttributes.RTSpecialName;
            writer.WriteAttributeString("isRTSpecialName", Utils.ilNS,
                b.ToString());
            b = mask == (int)FieldAttributes.HasFieldMarshal;
            writer.WriteAttributeString("isHasFieldMarshal", Utils.ilNS,
                b.ToString());
            b = mask == (int)FieldAttributes.HasDefault;
            writer.WriteAttributeString("isHasDefault", Utils.ilNS,
                b.ToString());
            b = mask == (int)FieldAttributes.HasFieldRVA;
            writer.WriteAttributeString("isHasFieldRVA", Utils.ilNS,
                b.ToString());

            #endregion

            foreach (RCustomAttrib attr in f.GetCustomAttributes(false)) {                
                AttributeAdapter.Instance.ToXml(attr, writer);
            }

            writer.WriteEndElement();
        }

        public void FromXml(RTypeDef t, XPathNavigator nav) {
            if (m_fields.Contains(
                nav.GetAttribute("definition", Utils.ilNS))) {

                return;
            }

            RType ft = TypeAdapter.Instance.GetType(
                nav.GetAttribute("type", Utils.ilNS));

			if (ft == null) throw new AssembleException(string.Format("Type {0} not found", nav.GetAttribute("type", Utils.ilNS)));


            RFieldDef f = t.DefineField(
                nav.GetAttribute("name", Utils.ilNS), ft);

            if (Utils.IsAttributePresent(nav, "value")) {
                f.Value = Utils.GetPrimitiveValue(
                    ft, nav.GetAttribute("value", Utils.ilNS));
            }

            m_fields[Utils.GetDefinition(f)] = f;

            #region FieldAttributes reading

            FieldAttributes attrs = new FieldAttributes();

            // Field access attributes
            if (Utils.IsAttributeTrue(nav, "isPrivateScope")) {
                attrs |= FieldAttributes.FieldAccessMask &
                    FieldAttributes.PrivateScope;
            }
            if (Utils.IsAttributeTrue(nav, "isPrivate")) {
                attrs |= FieldAttributes.FieldAccessMask &
                    FieldAttributes.Private;
            }
            if (Utils.IsAttributeTrue(nav, "isFamANDAssem")) {
                attrs |= FieldAttributes.FieldAccessMask &
                    FieldAttributes.FamANDAssem;
            }
            if (Utils.IsAttributeTrue(nav, "isAssembly")) {
                attrs |= FieldAttributes.FieldAccessMask &
                    FieldAttributes.Assembly;
            }
            if (Utils.IsAttributeTrue(nav, "isFamORAssem")) {
                attrs |= FieldAttributes.FieldAccessMask &
                    FieldAttributes.FamORAssem;
            }
            if (Utils.IsAttributeTrue(nav, "isPublic")) {
                attrs |= FieldAttributes.FieldAccessMask &
                    FieldAttributes.Public;
            }

            // Field contract attributes
            if (Utils.IsAttributeTrue(nav, "isStatic")) {
                attrs |= FieldAttributes.Static;
            }
            if (Utils.IsAttributeTrue(nav, "isInitOnly")) {
                attrs |= FieldAttributes.InitOnly;
            }
            if (Utils.IsAttributeTrue(nav, "isLiteral")) {
                attrs |= FieldAttributes.Literal;
            }
            if (Utils.IsAttributeTrue(nav, "isNotSerialized")) {
                attrs |= FieldAttributes.NotSerialized;
            }

            // Special name attribute
            if (Utils.IsAttributeTrue(nav, "isSpecialName")) {
                attrs |= FieldAttributes.SpecialName;
            }

            // Interop attributes
            if (Utils.IsAttributeTrue(nav, "isPinvokeIml")) {
                attrs |= FieldAttributes.PinvokeImpl;
            }

            // Runtime reserved attributes
            if (Utils.IsAttributeTrue(nav, "isRTSpecialName")) {
                attrs |= FieldAttributes.ReservedMask &
                    FieldAttributes.RTSpecialName;
            }
            if (Utils.IsAttributeTrue(nav, "isHasFieldMarshal")) {
                attrs |= FieldAttributes.ReservedMask &
                    FieldAttributes.HasFieldMarshal;
            }
            if (Utils.IsAttributeTrue(nav, "isHasDefault")) {
                attrs |= FieldAttributes.ReservedMask &
                    FieldAttributes.HasDefault;
            }
            if (Utils.IsAttributeTrue(nav, "isHasFieldRVA")) {
                attrs |= FieldAttributes.ReservedMask &
                    FieldAttributes.HasFieldRVA;
            }

            f.Attributes = attrs;

            #endregion

            XPathNodeIterator it = nav.Select(
                AttributeAdapter.Instance.TagName);
            while (it.MoveNext()) {
                AttributeAdapter.Instance.FromXml(f, it.Current);
            }
        }

        public RField GetField(string definition) {
			#region DEBUG
#if DEBUG
			Logger.Log("GetField({0})", definition);
#endif 
			#endregion
			RField result = (RField)m_fields[definition];				
            if (result == null) {
                RType t = TypeAdapter.Instance.GetType(
                    Utils.GetMemberTypeInSigOrDef(definition));
                foreach (RField f in t.GetFields()) {
                    if(Utils.GetDefinition(f) == definition) {
                        result = f;
                        break;
                    }
                }
                if (result != null) {
                    m_fields[definition] = result;
                }
            }
            if (result == null) {
                XPathNodeIterator it =
                    AssemblyAdapter.Instance.Navigator.Select(string.Format(
                    "//Field[@definition='{0}']", definition));
                while (it.MoveNext()) {
                    RTypeDef type = (RTypeDef)TypeAdapter.Instance.GetType(
                        Utils.GetParentFullName(it.Current));
                    FromXml(type, it.Current);
                    result = (RField)m_fields[definition];
                }
            }
            return result;
        }
    }
}
