/*
 * Copyright (c) 2004 DotNetGuru and the individuals listed
 * on the ChangeLog entries.
 *
 * Authors :
 *   Evain Jb   (jb@evain.net)
 *
 * This is a free software distributed under a BSD license
 * See LICENSE.BSD file for more details
 *
 *****************************************************************************/

namespace Ilml.Adapters {

	using System;
    using System.Reflection.Emit;
    using System.Text;
	using System.Xml;
	using System.Xml.XPath;

    using Rail.MSIL;
    using Rail.Reflect;

	internal sealed class InstructionAdapter : XmlAdapter {

        #region singleton implementation

		private InstructionAdapter() : base("Instruction") {}
		public static readonly InstructionAdapter Instance = 
			new InstructionAdapter();

        #endregion

        public void ToXml(CodeNodeInstruction ins, XmlWriter writer) {
            writer.WriteStartElement(TagName);
            Instruction i = ins.Instruction;
            writer.WriteAttributeString("offset", i.Index.ToString());
            writer.WriteAttributeString("code", i.OpCode.ToString());

            if (!(i is ILNone)) {
                string operand = null;
                if (i is ILBranch) {
                    operand = ((ILBranch)i).Target.Index.ToString();
                } else if (i is ILLocalVariable) {
                    if (i.OpCode.OperandType == OperandType.InlineVar ||
                        i.OpCode.OperandType == OperandType.ShortInlineVar) {

                        ILLocalVariable l = (ILLocalVariable)i;
                        for (int j = 0 ; j < l.Local.Method.LocalVarsCount
                            ; j++) {

                            if (l.Local.Method.GetLocalVar(j) == l.Local) {
                                operand = j.ToString(); break;
                            }
                        }
                    }
                } else if (i is ILArgument) {
                    if (i.OpCode.OperandType == OperandType.InlineVar ||
                        i.OpCode.OperandType == OperandType.ShortInlineVar) {

                        ILArgument a = (ILArgument)i;
                        operand = a.Parameter.Sequence.ToString();
                    }
                } else if (i is ILMethod) {
                    operand = Utils.GetSignature(((ILMethod)i).Method);
                } else if (i is ILField) {
                    operand = Utils.GetDefinition(((ILField)i).Field);
                } else if (i is ILOperandI2) {
                    operand = ((ILOperandI2)i).Operand.ToString();
                } else if (i is ILType) {
                    operand = Utils.GetFullyQualifiedName(((ILType)i).Type);
                } else if (i is ILToken) {
                    object op = ((ILToken)i).Token;
                    if (op is RMethodBase) {
                        operand = Utils.GetSignature((RMethodBase)op);
                    } else if (op is RField) {
                        operand = Utils.GetDefinition((RField)op);
                    } else if (op is RType) {
                        operand = Utils.GetFullyQualifiedName((RType)op);
                    }
                } else if (i is ILSig) {
                    throw new NotImplementedException();
                } else if (i is ILString) {
                    operand = ((ILString)i).String;
                } else if (i is ILOperandI4) {
                    operand = ((ILOperandI4)i).Operand.ToString();
                } else if (i is ILOperandI8) {
                    operand = ((ILOperandI8)i).Operand.ToString();
                } else if (i is ILOperandR4) {
                    operand = ((ILOperandR4)i).Operand.ToString();
                } else if (i is ILOperandR8) {
                    operand = ((ILOperandR8)i).Operand.ToString();
                } else if (i is ILSwitch) {
                    ILSwitch sw = (ILSwitch)i;
                    if (sw.Targets != null && sw.Targets.Length > 0) {
                        StringBuilder sb = new StringBuilder();
                        for (int j = 0 ; j < sw.Targets.Length ; j++) {
                            sb.Append(sw.Targets[j].Index);
                            if (j < sw.Targets.Length - 1) {
                                sb.Append(",");
                            }
                        }
                        operand = sb.ToString();
                    }
                }
                if (operand != null) {
                    writer.WriteAttributeString("operand", operand);
                }
            }
            writer.WriteEndElement();
        }

        public void FromXml(MethodBody mb, CodeNodeInstruction ins) {
            Instruction i = ins.Instruction;

            if (ins.Operand != null) {
                if (i is ILBranch) {
					CodeTreeXml tr = (CodeTreeXml)ins.Tree;
                    ILBranch br = (ILBranch)i;
					br.Target = tr.GetInstruction(
						tr.GetIndexForIdentifier(ins.Operand));
                } else if (i is ILLocalVariable) {
                    ((ILLocalVariable)i).Local = mb.GetLocalVar(
                    int.Parse(ins.Operand));
                } else if (i is ILArgument) {
                    ((ILArgument)i).Parameter = mb.RMethod.GetParameter(
                        int.Parse(ins.Operand));
                } else if (i is ILMethod) {
                    ILMethod meth = (ILMethod)i;
                    RMethodBase m = MethodAdapter.Instance.GetMethod(
                        ins.Operand);
                    if (m == null) {
                        m = ConstructorAdapter.Instance.GetConstructor(
                            ins.Operand);
                    }
                    if (m == null) {
                        throw new AssembleException(string.Format(
                            "Method {0} not found", ins.Operand));
                    }
                    meth.Method = m;
                } else if (i is ILField) {
                    RField field = FieldAdapter.Instance.GetField(
                        ins.Operand);
                    if (field == null) {
                        throw new AssembleException(string.Format(
                            "Field {0} not found", ins.Operand));
                    }
                    ((ILField)i).Field = field;
                } else if (i is ILOperandI2) {
                    ((ILOperandI2)i).Operand = short.Parse(ins.Operand);
                } else if (i is ILType) {
                    RType type = TypeAdapter.Instance.GetType(ins.Operand);
                    if (type == null) {
                        throw new AssembleException(string.Format(
                            "Type {0} not found", ins.Operand));
                    }
                    ((ILType)i).Type = type;
                        
                } else if (i is ILToken) {
                    RMember op = TypeAdapter.Instance.GetType(ins.Operand);
                    if (op == null) {
                        op = MethodAdapter.Instance.GetMethod(ins.Operand);
                    }
                    if (op == null) {
                        op = ConstructorAdapter.Instance.GetConstructor(
                            ins.Operand);
                    }
                    if (op == null) {
                        op = FieldAdapter.Instance.GetField(
                            ins.Operand);
                    }
                    if (op == null) {
                        throw new AssembleException(string.Format(
                            "Token {0} not found", ins.Operand));
                    }
                    ((ILToken)i).Token = op;
                } else if (i is ILString) {
                    ((ILString)i).String = ins.Operand;
                } else if (i is ILOperandI4) {
                    ((ILOperandI4)i).Operand = int.Parse(ins.Operand);
                } else if (i is ILOperandI8) {
                    ((ILOperandI8)i).Operand = long.Parse(ins.Operand);
                } else if (i is ILOperandR4) {
                    ((ILOperandR4)i).Operand = float.Parse(ins.Operand);
                } else if (i is ILOperandR8) {
                    ((ILOperandR8)i).Operand = double.Parse(ins.Operand);
                } else if (i is ILSwitch) {
					CodeTreeXml tr = (CodeTreeXml)ins.Tree;
                    ILSwitch sw = (ILSwitch)i;
                    string[] tokens = ins.Operand.Split(new char[] {','});
                    Instruction[] targets = new Instruction[tokens.Length];
                    for (int j = 0 ; j < tokens.Length ; j++) {
						targets[j] = tr.GetInstruction(
							tr.GetIndexForIdentifier(tokens[j]));
                    }
                    sw.Targets = targets;
                }
            }
            mb.Code.Add(i);
        }
	}
}
