/*
 * Copyright (c) 2004 DotNetGuru and the individuals listed
 * on the ChangeLog entries.
 *
 * Authors :
 *   Thomas Gil (thomas.gil@dotnetguru.org)
 *   Evain Jb   (jb@evain.net)
 *
 * This is a free software distributed under a BSD license
 * See LICENSE.BSD file for more details
 *
 *****************************************************************************/

namespace Ilml.Adapters {

    using System;
    using System.Reflection;
    using System.Xml;
    using System.Xml.XPath;

    using Rail.Reflect;
    
    internal sealed class PropertyAdapter : XmlAdapter {

        #region singleton implementation

        private PropertyAdapter() : base("Property") {}
        public static readonly PropertyAdapter Instance = 
            new PropertyAdapter();

        #endregion

        public void ToXml(RPropertyDef p, XmlWriter writer) {			
            writer.WriteStartElement(TagName, Utils.ilNS);

            writer.WriteAttributeString("name", Utils.ilNS, p.Name);
            writer.WriteAttributeString("type", Utils.ilNS, 
                Utils.GetFullyQualifiedName(p.PropertyType));
            writer.WriteAttributeString("definition", Utils.ilNS,
                Utils.GetDefinition(p));

            writer.WriteAttributeString("readable", Utils.ilNS,
                p.CanRead.ToString());
            writer.WriteAttributeString("writable", Utils.ilNS,
                p.CanWrite.ToString());

            if (p.CanRead) {
                writer.WriteAttributeString("getMethodSignature", 
                    Utils.ilNS, Utils.GetSignature(p.GetMethod));
            }
            if (p.CanWrite) {
                writer.WriteAttributeString("setMethodSignature", 
                    Utils.ilNS, Utils.GetSignature(p.SetMethod));
            }

            #region PropertyAttributes writing

            bool b;

            // Generals attributes
            b = (p.Attributes & PropertyAttributes.SpecialName) ==
                PropertyAttributes.SpecialName;
            writer.WriteAttributeString("isSpecialName", Utils.ilNS,
                b.ToString());
            b = p.Attributes == PropertyAttributes.None;
            writer.WriteAttributeString("isNone", Utils.ilNS,
                b.ToString());

            // Runtime reserved attributes
            int mask = (int)(p.Attributes & PropertyAttributes.ReservedMask);
            b = mask == (int)PropertyAttributes.RTSpecialName;
            writer.WriteAttributeString("isRTSpecialName", Utils.ilNS,
                b.ToString());
            b = mask == (int)PropertyAttributes.HasDefault;
            writer.WriteAttributeString("isHasDefault", Utils.ilNS,
                b.ToString());
            b = mask == (int)PropertyAttributes.Reserved2;
            writer.WriteAttributeString("isReserved2", Utils.ilNS,
                b.ToString());
            b = mask == (int)PropertyAttributes.Reserved3;
            writer.WriteAttributeString("isReserved3", Utils.ilNS,
                b.ToString());
            b = mask == (int)PropertyAttributes.Reserved4;
            writer.WriteAttributeString("isReserved4", Utils.ilNS,
                b.ToString());

            #endregion


            foreach (RCustomAttrib attr in p.GetCustomAttributes(false)) {
                AttributeAdapter.Instance.ToXml(attr, writer);
            }

            foreach (RParameter param in p.GetIndexParameters()) {
                ParameterAdapter.Instance.ToXml(param, writer);
            }
						
            writer.WriteEndElement();
        }

        public void FromXml(RTypeDef t, XPathNavigator nav) {
            RPropertyDef p = t.DefineProperty(
                nav.GetAttribute("name", Utils.ilNS),
                TypeAdapter.Instance.GetType(
                    nav.GetAttribute("type", Utils.ilNS)));

            p.SetIndexParameters(ParameterAdapter.Instance.FromXml(
                nav.Select(ParameterAdapter.Instance.TagName)));

            if (Utils.IsAttributeTrue(nav, "readable")) {
                p.GetMethod = MethodAdapter.Instance.GetMethod(
                    nav.GetAttribute("getMethodSignature", Utils.ilNS));
            }

            if (Utils.IsAttributeTrue(nav, "writable")) {
                p.SetMethod = MethodAdapter.Instance.GetMethod(
                    nav.GetAttribute("setMethodSignature", Utils.ilNS));
            }

            #region PropertyAttributes reading

            PropertyAttributes attrs = new PropertyAttributes();

            // Generals attributes
            if (Utils.IsAttributeTrue(nav, "isSpecialName")) {
                attrs |= PropertyAttributes.SpecialName;
            }

            // Runtime reserved attributes
            if (Utils.IsAttributeTrue(nav, "isRTSpecialName")) {
                attrs |= PropertyAttributes.ReservedMask &
                    PropertyAttributes.RTSpecialName;
            }
            if (Utils.IsAttributeTrue(nav, "isHasDefault")) {
                attrs |= PropertyAttributes.ReservedMask &
                    PropertyAttributes.HasDefault;
            }
            if (Utils.IsAttributeTrue(nav, "isReserved2")) {
                attrs |= PropertyAttributes.ReservedMask &
                    PropertyAttributes.Reserved2;
            }
            if (Utils.IsAttributeTrue(nav, "isReserved3")) {
                attrs |= PropertyAttributes.ReservedMask &
                    PropertyAttributes.Reserved3;
            }
            if (Utils.IsAttributeTrue(nav, "isReserved4")) {
                attrs |= PropertyAttributes.ReservedMask &
                    PropertyAttributes.Reserved4;
            }

            p.Attributes = attrs;

            #endregion

            XPathNodeIterator it = nav.Select(
                AttributeAdapter.Instance.TagName);
            while (it.MoveNext()) {
                AttributeAdapter.Instance.FromXml(p, it.Current);
            }
        }
    }
}
