/*
 * Copyright (c) 2004 DotNetGuru and the individuals listed
 * on the ChangeLog entries.
 *
 * Authors :
 *   Thomas Gil (thomas.gil@dotnetguru.org)
 *   Evain Jb   (jb@evain.net)
 *
 * This is a free software distributed under a BSD license
 * See LICENSE.BSD file for more details
 *
 *****************************************************************************/

namespace Ilml.Adapters {

    using System;
    using System.Collections;
    using System.Reflection;
    using System.Xml;
    using System.Xml.XPath;

    using Rail.Reflect;

    internal sealed class TypeAdapter : XmlAdapter {

        #region singleton implementation

		private TypeAdapter() : base("Type") {
            m_types = new Hashtable();
        }
		public static readonly TypeAdapter Instance = 
			new TypeAdapter();

        #endregion

		private IDictionary m_types;

		public bool IsTypeDefined(RType t) {
			return m_types.Contains(t.FullName);	
		}

		public void ToXml(RTypeDef t, XmlWriter writer) {
			writer.WriteStartElement(TagName, Utils.ilNS);

			if (t.BaseType != null) {
				writer.WriteAttributeString("baseType",
                    Utils.ilNS,
                    Utils.GetFullyQualifiedName(t.BaseType));
			}
			if (t.DeclaringType != null) {
				writer.WriteAttributeString("declaringType",
                    Utils.ilNS,
                    Utils.GetFullyQualifiedName(t.DeclaringType));
			}

            if (t.ElementType != null) {
                writer.WriteAttributeString("elementType",
                    Utils.ilNS,
                    Utils.GetFullyQualifiedName(t.ElementType));
            }

			writer.WriteAttributeString("fullName", Utils.ilNS,
                Utils.GetFullyQualifiedName(t));
			writer.WriteAttributeString("name", Utils.ilNS, t.Name);
			writer.WriteAttributeString("namespace", Utils.ilNS, t.NameSpace);

            #region TypeAttributes writing

            bool b;

            // Visibility attributes
            int mask = (int)(t.Attributes & TypeAttributes.VisibilityMask);
            b = mask == (int)TypeAttributes.NotPublic;
            writer.WriteAttributeString("isNotPublic", Utils.ilNS,
                b.ToString());
            b = mask == (int)TypeAttributes.Public;
            writer.WriteAttributeString("isPublic", Utils.ilNS,
                b.ToString());
            b = mask == (int)TypeAttributes.NestedPublic;
            writer.WriteAttributeString("isNestedPublic", Utils.ilNS,
                b.ToString());
            b = mask == (int)TypeAttributes.NestedPrivate;
            writer.WriteAttributeString("isNestedPrivate", Utils.ilNS,
                b.ToString());
            b = mask == (int)TypeAttributes.NestedFamily;
            writer.WriteAttributeString("isNestedFamily", Utils.ilNS,
                b.ToString());
            b = mask == (int)TypeAttributes.NestedFamANDAssem;
            writer.WriteAttributeString("isNestedFamANDAssem", Utils.ilNS,
                b.ToString());
            b = mask == (int)TypeAttributes.NestedFamORAssem;
            writer.WriteAttributeString("isNestedFamORAssem", Utils.ilNS,
                b.ToString());
            
            // Layout attributes
            mask = (int)(t.Attributes & TypeAttributes.LayoutMask);
            b = mask == (int)TypeAttributes.AutoLayout;
            writer.WriteAttributeString("isAutoLayout", Utils.ilNS,
                b.ToString());
            b = mask == (int)TypeAttributes.SequentialLayout;
            writer.WriteAttributeString("isSequencialLayout", Utils.ilNS,
                b.ToString());
            b = mask == (int)TypeAttributes.ExplicitLayout;
            writer.WriteAttributeString("isExplicitLayout", Utils.ilNS,
                b.ToString());

            // Class semantic attributes
            mask = (int)(t.Attributes & TypeAttributes.ClassSemanticsMask);
            b = mask == (int)TypeAttributes.Class;
            writer.WriteAttributeString("isClass", Utils.ilNS,
                b.ToString());
            b = mask == (int)TypeAttributes.Interface;
            writer.WriteAttributeString("isInterface", Utils.ilNS,
                b.ToString());

            // Special semantic attributes
            b = (t.Attributes & TypeAttributes.Abstract) ==
                TypeAttributes.Abstract;
            writer.WriteAttributeString("isAbstract", Utils.ilNS,
                b.ToString());
            b = (t.Attributes & TypeAttributes.Sealed) ==
                TypeAttributes.Sealed;
            writer.WriteAttributeString("isSealed", Utils.ilNS,
                b.ToString());
            b = (t.Attributes & TypeAttributes.SpecialName) ==
                TypeAttributes.SpecialName;
            writer.WriteAttributeString("isSpecialName", Utils.ilNS,
                b.ToString());

            // Implementation attributes
            b = (t.Attributes & TypeAttributes.Import) ==
                TypeAttributes.Import;
            writer.WriteAttributeString("isImport", Utils.ilNS,
                b.ToString());
            b = (t.Attributes & TypeAttributes.Serializable) ==
                TypeAttributes.Serializable;
            writer.WriteAttributeString("isSerializable", Utils.ilNS,
                b.ToString());

            // String Format attributes
            mask = (int)(t.Attributes & TypeAttributes.StringFormatMask);
            b = mask == (int)TypeAttributes.AnsiClass;
            writer.WriteAttributeString("isAnsiClass", Utils.ilNS,
                b.ToString());
            b = mask == (int)TypeAttributes.UnicodeClass;
            writer.WriteAttributeString("isUnicodeClass", Utils.ilNS,
                b.ToString());
            b = mask == (int)TypeAttributes.AutoClass;
            writer.WriteAttributeString("isAutoClass", Utils.ilNS,
                b.ToString());

            // Before field init
            b = (t.Attributes & TypeAttributes.BeforeFieldInit) ==
                TypeAttributes.BeforeFieldInit;
            writer.WriteAttributeString("isBeforeFieldInit", Utils.ilNS,
                b.ToString());

            // Runtime Reserved attributes
            mask = (int)(t.Attributes & TypeAttributes.ReservedMask);
            b = mask == (int)TypeAttributes.RTSpecialName;
            writer.WriteAttributeString("isRTSpecialName", Utils.ilNS,
                b.ToString());
            b = mask == (int)TypeAttributes.HasSecurity;
            writer.WriteAttributeString("isHasSecurity", Utils.ilNS,
                b.ToString());

            #endregion

            foreach (RType interf in t.GetInterfaces()) {
                writer.WriteElementString("ImplementsInterface", Utils.ilNS,
                    Utils.GetFullyQualifiedName(interf));
            }
            foreach (RCustomAttrib attr in t.GetCustomAttributes(false)) {
                AttributeAdapter.Instance.ToXml(attr, writer);
            }
			foreach (RFieldDef f in t.GetFields()) {
				FieldAdapter.Instance.ToXml(f, writer);
            }
            foreach (RPropertyDef p in t.GetProperties()) {
                PropertyAdapter.Instance.ToXml(p, writer);
            }
			foreach (RConstructorDef c in t.GetConstructors()) {
				ConstructorAdapter.Instance.ToXml(c, writer);
			}
			foreach (RMethodDef m in t.GetMethods()) {
				MethodAdapter.Instance.ToXml(m, writer);
			}
            foreach (REventDef e in t.GetEvents()) {
                EventAdapter.Instance.ToXml(e, writer);
            }

			writer.WriteEndElement();
		}

        public void FromXml(RModuleDef mod, XPathNavigator nav) {
            if (m_types.Contains(nav.GetAttribute("fullName", Utils.ilNS))) {
                return;
            }

            RTypeDef t = mod.DefineType(
                nav.GetAttribute("name", Utils.ilNS),
                nav.GetAttribute("namespace", Utils.ilNS));

            if (Utils.IsAttributePresent(nav, "baseType")) {
                t.BaseType = GetType(
                    nav.GetAttribute("baseType", Utils.ilNS));
            }

            if (Utils.IsAttributePresent(nav, "declaringType")) {
                t.DeclaringType = GetType(
                    nav.GetAttribute("declaringType", Utils.ilNS));
            }

            #region TypeAttributes reading

            TypeAttributes attrs = new TypeAttributes();

            // Visibility attributes
            if (Utils.IsAttributeTrue(nav, "isNotPublic")) {
                attrs |= TypeAttributes.VisibilityMask &
                    TypeAttributes.NotPublic;
            }
            if (Utils.IsAttributeTrue(nav, "isPublic")) {
                attrs |= TypeAttributes.VisibilityMask &
                    TypeAttributes.Public;
            }
            if (Utils.IsAttributeTrue(nav, "isNotPublic")) {
                attrs |= TypeAttributes.VisibilityMask &
                    TypeAttributes.NotPublic;
            }
            if (Utils.IsAttributeTrue(nav, "isNestedPublic")) {
                attrs |= TypeAttributes.VisibilityMask &
                    TypeAttributes.NestedPublic;
            }
            if (Utils.IsAttributeTrue(nav, "isNestedPrivate")) {
                attrs |= TypeAttributes.VisibilityMask &
                    TypeAttributes.NestedPrivate;
            }
            if (Utils.IsAttributeTrue(nav, "isNestedFamily")) {
                attrs |= TypeAttributes.VisibilityMask &
                    TypeAttributes.NestedFamily;
            }
            if (Utils.IsAttributeTrue(nav, "isNestedFamANDAssem")) {
                attrs |= TypeAttributes.VisibilityMask &
                    TypeAttributes.NestedFamANDAssem;
            }
            if (Utils.IsAttributeTrue(nav, "isNestedFamORAssem")) {
                attrs |= TypeAttributes.VisibilityMask &
                    TypeAttributes.NestedFamORAssem;
            }
            
            // Layout attributes
            if (Utils.IsAttributeTrue(nav, "isAutoLayout")) {
                attrs |= TypeAttributes.LayoutMask &
                    TypeAttributes.AutoLayout;
            }
            if (Utils.IsAttributeTrue(nav, "isSequencialLayout")) {
                attrs |= TypeAttributes.LayoutMask &
                    TypeAttributes.SequentialLayout;
            }
            if (Utils.IsAttributeTrue(nav, "isExplicitLayout")) {
                attrs |= TypeAttributes.LayoutMask &
                    TypeAttributes.ExplicitLayout;
            }

            // Class semantic attributes
            if (Utils.IsAttributeTrue(nav, "isClass")) {
                attrs |= TypeAttributes.ClassSemanticsMask &
                    TypeAttributes.Class;
            }
            if (Utils.IsAttributeTrue(nav, "isInterface")) {
                attrs |= TypeAttributes.ClassSemanticsMask &
                    TypeAttributes.Interface;
            }

            // Special semantic attributes
            if (Utils.IsAttributeTrue(nav, "isAbstract")) {
                attrs |= TypeAttributes.Abstract;
            }
            if (Utils.IsAttributeTrue(nav, "isSealed")) {
                attrs |= TypeAttributes.Sealed;
            }
            if (Utils.IsAttributeTrue(nav, "isSpecialName")) {
                attrs |= TypeAttributes.SpecialName;
            }

            // Implementation attributes
            if (Utils.IsAttributeTrue(nav, "isImport")) {
                attrs |= TypeAttributes.Import;
            }
            if (Utils.IsAttributeTrue(nav, "isSerializable")) {
                attrs |= TypeAttributes.Serializable;
            }

            // String Format attributes
            if (Utils.IsAttributeTrue(nav, "isAnsiClass")) {
                attrs |= TypeAttributes.StringFormatMask &
                    TypeAttributes.AnsiClass;
            }
            if (Utils.IsAttributeTrue(nav, "isUnicodeClass")) {
                attrs |= TypeAttributes.StringFormatMask &
                    TypeAttributes.UnicodeClass;
            }
            if (Utils.IsAttributeTrue(nav, "isAutoClass")) {
                attrs |= TypeAttributes.StringFormatMask &
                    TypeAttributes.AutoClass;
            }

            // Before field init
            if (Utils.IsAttributeTrue(nav, "isBeforeFieldInit")) {
                attrs |= TypeAttributes.BeforeFieldInit;
            }

            // Runtime Reserved attributes
            if (Utils.IsAttributeTrue(nav, "isRTSpecialName")) {
                attrs |= TypeAttributes.ReservedMask &
                    TypeAttributes.RTSpecialName;
            }
            if (Utils.IsAttributeTrue(nav, "isHasSecurity")) {
                attrs |= TypeAttributes.ReservedMask &
                    TypeAttributes.HasSecurity;
            }

            t.Attributes = attrs;

            #endregion

            m_types.Add(Utils.GetFullyQualifiedName(t), t);

            XPathNodeIterator it = nav.Select("ImplementsInterface");
            while (it.MoveNext()) {
                t.AddInterface(GetType(it.Current.Value));
            }

            it = nav.Select(AttributeAdapter.Instance.TagName);
            while (it.MoveNext()) {
                AttributeAdapter.Instance.FromXml(t, it.Current);
            }
            it = nav.Select(FieldAdapter.Instance.TagName);
            while (it.MoveNext()) {
                FieldAdapter.Instance.FromXml(t, it.Current);
            }
            it = nav.Select(PropertyAdapter.Instance.TagName);
            while (it.MoveNext()) {
                PropertyAdapter.Instance.FromXml(t, it.Current);
            }
            it = nav.Select(ConstructorAdapter.Instance.TagName);
            while (it.MoveNext()) {
                ConstructorAdapter.Instance.FromXml(t, it.Current);
            }
            it = nav.Select(MethodAdapter.Instance.TagName);
            while (it.MoveNext()) {
                MethodAdapter.Instance.FromXml(t, it.Current);
            }
            it = nav.Select(EventAdapter.Instance.TagName);
            while (it.MoveNext()) {
                EventAdapter.Instance.FromXml(t, it.Current);
            }
        }
        
        public RType GetType(string fullName) {
			#region DEBUG
#if DEBUG
			Logger.Log("GetType({0})", fullName);
#endif 
			#endregion
			RType t = (RType)m_types[fullName];
            if (t == null && fullName.IndexOf("[]") > -1) {
                string name = fullName.Substring(0, fullName.IndexOf("[]"));
                RType elem = (RType)m_types[name];
                if (elem == null) {
                    GetType(name);
                }
            }
            if (t == null && !fullName.StartsWith("<PrivateImplementationDetails>")) {
                t = AssemblyAdapter.Instance.CurrentAssembly.GetRTypeRef(
                    fullName);
                if (t == null) {
                    foreach (RModule mod in ModuleAdapter.Instance.Modules.Values) {
                        if ((t = mod.GetType(fullName)) != null) {
                            break;
                        }
                    }
                }
                if (t != null) {
                    m_types[Utils.GetFullyQualifiedName(t)] = t;
                }
            }
            if (t == null) {
                XPathNodeIterator it =
                    AssemblyAdapter.Instance.Navigator.Select(string.Format(
                    "//Type[@fullName='{0}']", fullName));
                while (it.MoveNext()) {
                    RModuleDef mod = ModuleAdapter.Instance.GetModule(
                        Utils.GetParentFullName(it.Current));
                    FromXml(mod, it.Current);
                    t = (RType)m_types[fullName];
                }
            }
            return t;
        }
	}
}
