/*
 * Copyright (c) 2004 DotNetGuru and the individuals listed
 * on the ChangeLog entries.
 *
 * Authors :
 *   Thomas Gil (thomas.gil@dotnetguru.org)
 *   Evain Jb   (jb@evain.net)
 *
 * This is a free software distributed under a BSD license
 * See LICENSE.BSD file for more details
 *
 *****************************************************************************/

namespace Ilml.Adapters {

    using System;
    using System.Collections;
    using System.Text;
    using System.Xml;
    using System.Xml.XPath;

    using Rail.MSIL;
    using Rail.Reflect;

    internal sealed class Utils  {

        private Utils() {}

        public static string GetFullyQualifiedName(RType t) {
            string ret = null;
            if (t.DeclaringType != null) {
                ret = string.Format("{0}+{1}",
                    GetFullyQualifiedName(t.DeclaringType), t.Name);
            } else if (t.NameSpace == null || (t.NameSpace != null && t.NameSpace.Length == 0)) {
                ret = t.Name;
            } else {
                ret = string.Format("{0}.{1}", t.NameSpace, t.Name);
            }
            return ret;
        }

		// Signatures look like: 
		// - "ReturnType OperationName(FullyQualifiedParamType,etc...)
        public static string GetSignature(RMethodBase meth) {
            StringBuilder sb = new StringBuilder();
            if (meth is RMethod) {
				RMethod m = (RMethod)meth;
                sb.Append(GetFullyQualifiedName(m.ReturnType.ReturnType));
			} else {
                sb.Append("System.Void");
            }
            sb.Append(" ");
            GetMemberDefinition(meth, sb);
            sb.Append("(");

            RParameter[] parameters = meth.GetParameters();
            for (int i = 0 ; i < parameters.Length ; i++) {
                sb.Append(GetFullyQualifiedName(
                    parameters[i].ParameterType));
                if (i < parameters.Length - 1) {
                    sb.Append(",");
                }
            }
            sb.Append(")");

            return sb.ToString();
        }

        public static string GetDefinition(RField field) {	
            StringBuilder sb = new StringBuilder();
            sb.Append(GetFullyQualifiedName(field.FieldType));
            sb.Append(" ");
            GetMemberDefinition(field, sb);
            return sb.ToString();
        }

        public static string GetDefinition(RProperty property) {
            StringBuilder sb = new StringBuilder();
            sb.Append(GetFullyQualifiedName(property.PropertyType));
            sb.Append(" ");
            GetMemberDefinition(property, sb);
            return sb.ToString();
        }

		// A member definition looks like : 
		// - "FullyQualifiedDeclaringType::MemberName"
        private static void GetMemberDefinition(RMember member, StringBuilder sb) {
            sb.Append(GetFullyQualifiedName(member.DeclaringType));
            sb.Append("::");
            sb.Append(member.Name);
		}

        public static string GetMemberTypeInSigOrDef(string sigdef) {
            int spacePos = sigdef.IndexOf(" ");
            return sigdef.Substring(spacePos + 1,
                sigdef.IndexOf("::") - spacePos - 1);
        }

        public static object GetPrimitiveValue(RType t, string val) {
            string name = (t.BaseType != null &&
                Utils.GetFullyQualifiedName(t.BaseType) == "System.Enum") ?
                "System.Enum" : Utils.GetFullyQualifiedName(t);
            try {
                switch (name) {
                    case "System.String" :
                        return val;
                    case "System.Int16" :
                        return Int16.Parse(val);
                    case "System.Int32" :
                        return Int32.Parse(val);
                    case "System.Int64" :
                        return Int64.Parse(val);
                    case "System.Byte" :
                        return Byte.Parse(val);
                    case "System.Char" :
                        return Char.Parse(val);
                    case "System.Double" :
                        return Double.Parse(val);
                    case "System.Decimal" :
                        return Decimal.Parse(val);
                    case "System.SByte" :
                        return SByte.Parse(val);
                    case "System.UInt16" :
                        return UInt16.Parse(val);
                    case "System.UInt32" :
                        return UInt32.Parse(val);
                    case "System.UInt64" :
                        return UInt64.Parse(val);
                    case "System.Single" :
                        return Single.Parse(val);
                    case "System.Boolean" :
                        return Boolean.Parse(val);
                    case "System.Enum" :
                        return GetPrimitiveValue(
                            GetEnumValueType(t), val);
                }
            } catch {}
            return null;
        }

        private static RType GetEnumValueType(RType t) {
            RField f = t.GetField("value__");
            return f.FieldType;
        }

        public const string ilNS = "";

        public static string GetSignature(XPathNavigator nav) {
            return nav.GetAttribute("signature", ilNS);
        }

        public static string GetParentFullName(XPathNavigator nav) {
            XPathNodeIterator it = nav.SelectAncestors(
                XPathNodeType.Element, false);
            it.MoveNext();
            return it.Current.GetAttribute("fullName", ilNS);
        }

        public static string GetParentSignature(XPathNavigator nav) {
            XPathNodeIterator it = nav.SelectAncestors(
                XPathNodeType.Element, false);
            it.MoveNext();
            return it.Current.GetAttribute("signature", ilNS);
        }

        public static bool IsAttributePresent(XPathNavigator nav,
            string attributeName) {

            return nav.GetAttribute(attributeName, ilNS).Length > 0;			
        }

        public static bool IsAttributeTrue(XPathNavigator nav,
            string attributeName) {

            try {
                return bool.Parse(nav.GetAttribute(attributeName, ilNS));
            } catch {
                return false;
            }
        }
    }
}